"""SnapFuture is a special subclass of a Future that is returned from the methods in FSC

It is primarily used to attach additional properties to the future when storing them in a
pending list.
"""
from tornado.concurrent import Future


class SnapFuture(Future):
    """This class inherits from future, but also includes some additional values needed by SNAP Connect"""
    def __init__(self, is_multicast, expected_seq, response_addrs, callback_name):
        """Create a SnapFuture that is resolved when an RPC or Mcast RPC responds

        Args:
            is_multicast (bool): True if this is a multicast response
            expected_seq (int): Expected application sequence number in the response
            response_addrs (str | list[str]): Hexlified addresses we expect to respond to the RPC
            callback_name (str): Name of the RPC response
        """
        super(SnapFuture, self).__init__()

        self.is_multicast = is_multicast  # True if this is a future for a multicast RPC
        self.expected_seq = expected_seq  # Expected sequence number of response, or None if we don't expect one
        self.responses = {}  # Hang partial responses off of the future
        self.callback_name = callback_name

        if self.is_multicast:
            # If response_addrs are provided, must be a list
            if response_addrs and response_addrs != ['']:
                self.expected_responses = [a.lower() for a in response_addrs]
            else:
                # Multicast that expects any responses
                self.expected_responses = ''
        else:
            # Just the one
            self.expected_responses = response_addrs.lower()

    @property
    def expects_any(self):
        return self.expected_responses == ''

    def expects(self, addr):
        if self.expects_any:
            return True
        if self.is_multicast:
            return addr.lower() in self.expected_responses
        else:
            return addr.lower() == self.expected_responses

    def response_is_duplicate(self, addr):
        if self.expects_any:
            return False
        return addr in self.responses

    def remove_from_expected(self, addr):
        if self.expects_any:
            return
        self.expected_responses.remove(addr)

    def has_all_responses(self):
        if self.expects_any:
            return False
        if not self.expected_responses:
            return True
        return False

    def get_key(self):
        """Return a unique key for this future

        This is used as an index into the callback futures dictionary"""
        if self.is_multicast:
            return self.callback_name
        else:
            # RPCs need to be more unique than just the function name
            return "%s.%s" % (self.expected_responses, self.callback_name)
