# coding=utf-8
# (c) Copyright 2009-2016 Synapse Wireless, Inc.
"""
SNAPconnect Futures is a wrapper for SNAPconnect that returns futures for most common operations. This
allows you to write asynchronous code using coroutines.

See the readme for usage examples
"""
import apy
import logging
import binascii
import random
import re
from collections import defaultdict, OrderedDict
from datetime import timedelta
from snapconnect import snap

from tornado.concurrent import Future
from tornado.gen import coroutine, Return, with_timeout, TimeoutError
from tornado.ioloop import PeriodicCallback

from .snap_future import SnapFuture
from .packet_overhead import get_max_dmcast_targets
from .responses import MultiResult
from .serial_manager import FutureSerialManager
from .upgrade_manager import FutureUpgradeManager

DEFAULT_MCAST_TTL = 5
DEFAULT_MCAST_GROUP = 0x0001
DEFAULT_DMCAST_DELAY = 50  # Delay in milliseconds between dmcast responses

RPC_RETRIES = 3  # Number of times to attempt to send an RPC.  This is NOT related to retries from not hearing responses.
DEFAULT_RETRIES = 3  # Number of times to retry a callback_* when we don't hear a response before timeout.

log = logging.getLogger('snapconnect_futures')
log_rpc = logging.getLogger("snapconnect_futures.rpc")  # lower level debugger

# Message type enumerations
MESSAGE_TYPE_RPC = "rpc"
MESSAGE_TYPE_MCAST = "mcast"
MESSAGE_TYPE_DMCAST = "dmcast"

VALID_SCF_ADDR = re.compile(r'^([0-9a-f]{6})*$')

# STORY TIME!
#
# The InvalidSNAPAddressError and validate_addresses function were
# added to guard SCF internals:
# SCF operates on (and assumes) SNAP addresses are lowercase.
#
# We could have done the .lower() conversion on the user's behalf,
# but this would've resulted in the user doing things like:
#
# result = scf.dmcast_rpc(["ABC123"], "foo", "bar")
# (NB: at this point result.successful looks like {"abc123": "response"})
#
# if "ABC123" in result.successful:
#      ...
#
# and having successful RPCs look like failures because we
# returned the result as coming from "abc123", not "ABC123".
#
# We could've done the work internally to preserve the calling
# case, but that was deemed to be inexpedient, and potentially
# confusing.
#
# We decided it was better to be consistent, and explicitly
# enforce that consistency, rather than attempt to accommodate
# inconsistent addressing schemes.

class InvalidSNAPAddressError(Exception):
    """Exception raised when an invalid SNAP address is provided"""

    def __init__(self, address=None):
        self.invalid_address = address
        super(InvalidSNAPAddressError, self).__init__("Invalid SNAPconnect Futures address, expected ([0-9a-f]{6})* (note lowercase): %r" % address)


# STORY TIME 2
#
# The callback_dmcast_rpc logic will intelligently break up a target list
# that's too long to fit into a packet.
# If there's too many arguments (or they're too large) there's not enough
# room in the packet for *any* targets.
# In that case, this is the exception we'll raise.

class CouldNotFitTargetsError(Exception):
    """Exception raised when not enough room is left in a dmcast packet for the target addresses"""

    def __init__(self, targets=None, func=None, *args):
        super(CouldNotFitTargetsError, self).__init__("Error encoding DMCast RPC %s%r for targets %s, could not fit any targets in packet." % (func, args, targets))


def validate_addresses(addrs):
    """Verify that SNAP address(es) are valid.

    Valid SNAP addresses are 6 lowercase hex digits.

    Args:
        addrs (str | list[str]): SNAP address(es) to validate.

    Returns:
        The validated addresses
    """
    if isinstance(addrs, basestring):
        if not VALID_SCF_ADDR.match(addrs):
            raise InvalidSNAPAddressError(addrs)
        return addrs
    else:
        return [validate_addresses(a) for a in addrs]


def SnapConnectFuturesAuto(sc_license=None, sc_address=None, sc_scheduler=None):
    """If you are unfamiliar with SNAPconnect Futures, you may want to use this helper function to instantiate a
    ``SnapConnectFutures`` object with common defaults that abstracts away some of the necessary calls.

    Args:
        sc_license: SNAPconnect license to use
        sc_address: SNAPconnect address to use (Ex: “aabbcc”)
        sc_scheduler: apy scheduler to use

    Returns:
        An instance of `SnapConnectFutures` initialized with a simple configuration.

    Example:

        .. code-block:: python

            from snapconnect_futures import SnapConnectFuturesAuto
            from snapconnect import snap
            from tornado.gen import coroutine
            import tornado

            scf = SnapConnectFuturesAuto()

    """
    if sc_scheduler is None:
        sc_scheduler = apy.ioloop_scheduler.IOLoopScheduler.instance()

    comm = snap.Snap(scheduler=sc_scheduler, license_file=sc_license, addr=sc_address, funcs={})
    PeriodicCallback(comm.poll_internals, 5).start()
    scf = SnapConnectFutures(comm)
    return scf


class SnapConnectFutures(object):
    """Wraps a ``Snap`` instance and allows you to write asynchronous code using coroutines.
    """

    def __init__(self, snapconnect):
        """Wrap a SNAPconnect instance with the ability to return Futures.

        Args:
            snapconnect (snap.Snap): SNAPconnect instance to wrap

        Example:

            .. code-block:: python

                from snapconnect_futures import SnapConnectFutures
                from snapconnect import snap
                from tornado.gen import coroutine
                import apy
                import tornado

                # Required IO loop setup:
                # Create an apy IOLoopScheduler
                scheduler = apy.ioloop_scheduler.IOLoopScheduler.instance()

                # Create a SNAPconnect instance that uses the apy scheduler
                comm = snap.Snap(scheduler=scheduler, funcs={})

                # Setup tornado's IO loop by scheduling calls to SNAPconnect's poll_internals method
                tornado.ioloop.PeriodicCallback(comm.poll_internals, 5).start()

                # Create an SCF instance using an existing SNAPconnect instance
                # This registers all of our callbacks automatically:
                scf = SnapConnectFutures(comm)

        """
        self._snap = snapconnect

        self._unique = defaultdict(int)  # Store the next unique number for each function callback

        self._app_seq = random.randint(1, 255)  # Application level sequence number

        self._rpc_futures = {}  # RPC Sent ID -> Future
        self._callback_futures = {}  # Address.FunctionName -> Future

        # When True, we will immediately resolve RPCs without waiting for HOOK_RPC_SENT
        self._rpcs_resolve_immediately = False
        self._immediately_resolve_rpcs(False)  # By default, wait to resolve RPCs

        # Add hooks for opening and closing serial connections
        self.serial_manager = FutureSerialManager(self._snap)

        # Add manager for performing script and firmware upgrades
        self.upgrade_manager = FutureUpgradeManager(self._snap)

        self.next_message_sent = Future()  # Future that resolves as soon as the next message sent

    def _immediately_resolve_rpcs(self, resolve_immediately=True):
        """Set FSC to immediately resolve RPC futures without HOOK_RPC_SENT being called first.

        This is mostly useful during testing.

        Args:
            resolve_immediately (bool):  Turn off/on resolving RPCs immediately
        """
        self._rpcs_resolve_immediately = resolve_immediately

        if self._rpcs_resolve_immediately:
            if self._rpc_futures:
                log.critical("Existing RPC futures area already waiting. These will not be automatically resolved.")

            self._snap.packet_dispatcher.unregister_callback(snap.hooks.HOOK_RPC_SENT, self._on_rpc_sent)
        else:
            self._snap.packet_dispatcher.register_callback(snap.hooks.HOOK_RPC_SENT, self._on_rpc_sent)

    @staticmethod
    def get_default_mcast_ttl():
        """Return the default multicast TTL. Override in a subclass if your application changes this dynamically.

        Returns:
            Default multicast TTL
        """
        return DEFAULT_MCAST_TTL

    @staticmethod
    def get_default_mcast_group():
        """Return the default multicast group. Override in a subclass if your application changes this dynamically.

        Returns:
            Default multicast group
        """
        return DEFAULT_MCAST_GROUP

    @staticmethod
    def get_default_dmcast_delay():
        """Return the default directed multicast delay. Override in a subclass if your application changes this dynamically.

        Returns:
            Default directed multicast delay
        """
        return DEFAULT_DMCAST_DELAY

    def _message_sent(self):
        """Called every time a message is successfully sent"""
        self.next_message_sent.set_result(True)
        self.next_message_sent = Future()

    def _get_unique_name(self, name):
        """Return a unique version of the given name"""
        self._unique[name] += 1
        return "%s%s" % (name, self._unique[name])

    def _on_rpc_sent(self, rpc_id, success):
        try:
            self._rpc_futures[rpc_id].set_result(success)
            del self._rpc_futures[rpc_id]
            self._message_sent()
        except KeyError:
            log_rpc.info('Could not find an existing Future for RPC id %s', rpc_id)
            log_rpc.debug('Outstanding RPC futures: %r', self._rpc_futures.keys())

    def _get_next_seq(self):
        """Get the next application sequence number"""
        self._app_seq = (self._app_seq + 1) % 256

        if self._app_seq == 0:
            self._app_seq = 1

        return self._app_seq

    def rpc(self, addr, funcname, *args):
        """Send a unicast RPC with no retry, timeout, or callback handling.

        Args:
            addr (str): SNAP address (Ex: "aabbcc", must be lowercase)
            funcname (str): Function to call
            *args: Arguments for RPC

        Returns:
            Future: Resolves to True if SNAPconnect believes the packet was sent. Resolves to False, otherwise.

        Example:

            .. code-block:: python

                rpc_success = yield scf.rpc("aabbcc", "test_func", "arg1", "arg2")
                # => True

        """
        addr = validate_addresses(addr)
        future = Future()
        log_rpc.info("RPC: %s.%s%r", addr, funcname, args)

        rpc_id = self._snap.rpc(binascii.unhexlify(addr), funcname, *args)

        if self._rpcs_resolve_immediately:
            future.set_result(True)
            self._message_sent()
        else:
            self._rpc_futures[rpc_id] = future

        return future

    def mcast_rpc(self, funcname, *args, **kwargs):
        """Send a multicast RPC with no retry, timeout, or callback handling.

        Args:
            funcname (str): Function to call
            *args: Arguments for RPC
            **kwargs: Keyword arguments for SNAPconnect.  View Keyword Arguments for specific kwargs expected.

        Keyword Args:
            group (int): Multicast Group. If not provided or None, use the default group. Default: None
            ttl (int): Number of hops. If not provided or None, use the default ttl. Default: None

        Returns:
            Future: Resolves to True if SNAPconnect believes the packet was sent.
            Resolves to False, otherwise.

        Example:

            .. code-block:: python

                mcast_rpc_success = yield scf.mcast_rpc("test_func", "arg1", "arg2")

        Note:
            * ``mcast_rpc`` uses :meth:`~SnapConnectFutures.get_default_mcast_group` to obtain the default mcast group *unless overridden* via kwargs
            * ``mcast_rpc`` uses :meth:`~SnapConnectFutures.get_default_mcast_ttl` to obtain the default mcast ttl *unless overridden* via kwargs

        """
        future = Future()

        ttl = kwargs.get("ttl", None)
        if ttl is None:
            ttl = self.get_default_mcast_ttl()

        group = kwargs.get("group", None)
        if group is None:
            group = self.get_default_mcast_group()

        log_rpc.info("Mcast RPC: %s%r", funcname, args)
        rpc_id = self._snap.mcast_rpc(group, ttl, funcname, *args)

        if self._rpcs_resolve_immediately:
            future.set_result(True)
            self._message_sent()
        else:
            self._rpc_futures[rpc_id] = future

        return future

    def dmcast_rpc(self, addrs, funcname, *args, **kwargs):
        """Send a directed multicast RPC with no retry, timeout, or callback handling.

        Args:
            addrs (str | list[str]): List of SNAP addresses (Ex: ["aabbcc", "ccddee"], must be lowercase)
            funcname (str): Function to call
            *args: Arguments for RPC
            **kwargs:  Keyword arguments for SNAPconnect.  View Keyword Arguments for specific kwargs expected.

        Keyword Args:
            delay (int): Millisecond delay between dmcast responses.
                If not provided or None, use default delay. Default: None
            group (int): Multicast Group. If not provided or None, use default group. Default: None
            ttl (int): Number of hops. If not provided or None, use default ttl. Default: None

        Returns:
            Future: Resolves to True if SNAPconnect believes the packet was sent. Resolves to False, otherwise.

        Example:

            .. code-block:: python

                dmcast_rpc_success = yield scf.dmcast_rpc(["aabbcc", "ccddee"], "test_func", "arg1", "arg2")
                # => True

        Note:
            * ``dmcast_rpc`` uses :meth:`~SnapConnectFutures.get_default_mcast_group` to obtain the default mcast group *unless overridden* via kwargs
            * ``dmcast_rpc`` uses :meth:`~SnapConnectFutures.get_default_mcast_ttl` to obtain the default mcast ttl *unless overridden* via kwargs
            * ``dmcast_rpc`` uses :meth:`~SnapConnectFutures.get_default_dmcast_delay` to obtain the default dmcast delay *unless overridden* via kwargs

        """

        addrs = validate_addresses(addrs)

        future = Future()

        addr_string = self.get_addr_str(addrs)

        delay = kwargs.get("delay", None)
        if delay is None:
            delay = self.get_default_dmcast_delay()

        ttl = kwargs.get("ttl", None)
        if ttl is None:
            ttl = self.get_default_mcast_ttl()

        group = kwargs.get("group", None)
        if group is None:
            group = self.get_default_mcast_group()

        log_rpc.info("DMcast RPC: %s%r", funcname, args)
        rpc_id = self._snap.dmcast_rpc(addr_string, group, ttl, delay, funcname, *args)

        if self._rpcs_resolve_immediately:
            future.set_result(True)
            self._message_sent()
        else:
            self._rpc_futures[rpc_id] = future

        return future

    def _get_callback(self, funcname):
        """Return a future callback for the given function name

        Returns:
            callable
        """

        def snap_future_callback(*args):
            source_addr = binascii.hexlify(self._snap.rpc_source_addr())
            log.debug("%s callback %s%r received", source_addr, funcname, args)

            f = self._callback_futures.pop("%s.%s" % (source_addr, funcname), None)

            if not f:
                # Try to see if there's a non-address specific future waiting for multicast
                f = self._callback_futures.get(funcname, None)

            if f and f.running():
                # See if we're expecting sequence numbers on this future
                # If we are, the first 2 arguments need to match our local address and seq
                if f.expected_seq is not None:
                    if args[:2] != (self._snap.local_addr(), f.expected_seq):
                        log.error("Callback did not have expected addr %s and sequence number %s", binascii.hexlify(self._snap.local_addr()), f.expected_seq)
                        return
                    else:
                        # We can strip off the first 2 arguments if they were valid
                        # We don't need to know about sequence number and our own address at lower levels
                        args = args[2:]

                if f.is_multicast:
                    # See if we already heard this one
                    if f.response_is_duplicate(source_addr):
                        # If we're receiving a duplicate at the application level, make sure to log this behavior
                        log.error("Received duplicate mcast for %s from %s", funcname, source_addr)
                    elif not f.expects(source_addr):
                        log.error("Received unexpected mcast for %s from %s", funcname, source_addr)
                    else:
                        # If we have them all, resolve the future
                        f.responses[source_addr] = args
                        f.remove_from_expected(source_addr)

                        if f.has_all_responses():
                            del self._callback_futures[funcname]
                            self._snap.local_dispatcher.funcs.pop(funcname, None)
                            f.set_result(f.responses)
                else:
                    self._snap.local_dispatcher.funcs.pop(funcname, None)
                    f.set_result(args)

            else:
                log.warning("Received callback from %s for %s%r but no future available", source_addr, funcname, args)

        return snap_future_callback

    @staticmethod
    def get_addr_str(addrs):
        """Return a list of hexlified addresses for RPCs.

        Args:
            addrs (str | list[str]): List of one or more unhexlifed addresses. Ex: ["aabbcc", "112233"]

        Raises:
            TypeError: If an invalid address is provided

        Returns:
            Comma-separated str of provided addresses
        """
        try:
            if isinstance(addrs, basestring):
                return binascii.unhexlify(addrs)
            else:
                return ''.join([binascii.unhexlify(addr) for addr in addrs])
        except:
            raise TypeError("Invalid address string %r" % addrs)

    def _register_callback(self, funcname):
        """Register an RPC callback that can resolve futures

        Args:
            funcname (str): Name of function to resolve
        """
        current_function = self._snap.local_dispatcher.funcs.get(funcname, None)

        if current_function and current_function.__name__ != "snap_future_callback":
            log.error("SNAP Connect function %s already registered for %s and will be replaced",
                      current_function.__name__, funcname)

        self._snap.local_dispatcher.funcs[funcname] = self._get_callback(funcname)

    @coroutine
    def _rpc_with_retries(self, message_type, funcname, args, addrs, ttl, dmcast_delay, group=None):
        """Return a future that attempts to send any type of RPC multiple times

        Args:
            message_type (str): MESSAGE_TYPE_MCAST, MESSAGE_TYPE_DMCAST, or MESSAGE_TYPE_RPC
            funcname (str): Function to be called by the RPC
            args (list): Arguments passed to the RPC
            addrs (str | list[str]): Address or addresses to send the message to if applicable
            ttl (int): TTL if applicable
            dmcast_delay (int): DMCast delay if applicable
            group (int): Group if applicable

        Returns:
            bool: If the RPC could never be sent out the system, the Future resolves to False.
                If the RPC could have been sent, the Future resolves to True.
        """
        for rpc_retry in xrange(RPC_RETRIES + 1):
            if message_type == MESSAGE_TYPE_MCAST:
                rpc_success = yield self.mcast_rpc(*args, ttl=ttl, group=group)
            elif message_type == MESSAGE_TYPE_RPC:
                rpc_success = yield self.rpc(addrs, *args)
            elif message_type == MESSAGE_TYPE_DMCAST:
                rpc_success = yield self.dmcast_rpc(addrs, *args, ttl=ttl, group=group, delay=dmcast_delay)

            if not rpc_success:
                log.error("RPC failed to send for %s %s%r", addrs if message_type == MESSAGE_TYPE_RPC else message_type, funcname, args)
            else:
                raise Return(True)

        raise Return(False)

    @staticmethod
    def get_results_with_result_type(result_type, callback_result):
        """Return the responses that match a result type.

        Args:
            result_type (namedtuple): The type of callback expected
            callback_result (dict[str, tuple]): The result of the callback

        Returns:
            tuple[dict, dict]: The converted results and the dict of invalid responses that were too long or too short
        """
        invalid_responses = {}

        if result_type:
            # If a namedtuple was provided, convert tuple results to the named variety
            if isinstance(callback_result, dict):
                # Responses with too many or too few arguments when a callback_result is provided are treated as failures
                invalid_responses = {k: v for k, v in callback_result.iteritems() if len(v) != len(result_type._fields)}
                callback_result = {k: result_type(*v) if k not in invalid_responses else None for k, v in callback_result.iteritems()}
            else:
                callback_result = result_type(*callback_result) if len(callback_result) == len(result_type._fields) else None

        return callback_result, invalid_responses

    @coroutine
    def _callback_message(self, addrs, funcname, *args, **kwargs):  # flake8: noqa
        """Handles internal mechanisms for sending mcast, dmcast, and rpc messages

        Args:
            addrs (str | list): SNAP Address (Ex: "aabbcc") or list of SNAP addresses to expect responses from
            funcname (str): Name of function to call with RPC
            *args: Arguments to pass to the node
            **kwargs: See Keyword Args for definition of valid kwargs.

        Keyword Args:
            callback_name (str): If None, send the RPC with 'callback' to a random callback
                If set to a value, expect your RPC to generate a response to a certain function and do NOT use
                'callback' in the RPC
            dmcast_delay (int): millisecond delay between dmcast responses.  If not provided or None, use default group
            expect_seq (bool): True if a sequence number will be appended to the RPC args.
                Used in most lighting RPCs. (dmcast only)
            group (int): Multicast Group. If not provided or None, use default group
            message_type (str): MESSAGE_TYPE_MCAST, MESSAGE_TYPE_DMCAST, or MESSAGE_TYPE_RPC
            node_limit (int | None): The max number of nodes allowed in a dmcast
                If None, fit as many as possible. (dmcast only)
            result_type (namedtuple): Named tuple describing the format of the returned value
            retries (int): Number of times to retry the message after a timeout
            timeout (float): Number of seconds to wait before timing out
            ttl (int): Number of hops. If not provided or None, use default ttl
            validate_func (function): The function to apply to each result of the callback result. This function should
                take one RPC response argument (a tuple or dict of response tuples) and return True or False.

        Yields:
            Return varies depending on the message_type.

            future[return value]: For rpc.

            dict[addrs, return values]: For mcast_rpc.

            tuple[dict[addrs, return values], unprocessed nodes]: For dmcast_rpc
        """

        # Extract out kwargs/set default values
        timeout = kwargs.get("timeout", 2)
        retries = kwargs.get("retries", DEFAULT_RETRIES)
        result_type = kwargs.get("result_type", None)
        callback_name = kwargs.get("callback_name", None)
        message_type = kwargs.get("message_type", None)
        ttl = kwargs.get("ttl", None)
        dmcast_delay = kwargs.get("dmcast_delay", None)
        expect_seq = kwargs.get("expect_seq", None)
        node_limit = kwargs.get("node_limit", None)
        group = kwargs.get("group", None)
        validate_func = kwargs.get("validate_func", None)

        # True if we should use RPC the 'callback' function
        # False when we just expect the function to respond on its own
        use_callback_rpc = False

        # The future would resolve to the response RPCs arguments
        if callback_name is None:
            use_callback_rpc = True
            callback_name = self._get_unique_name(funcname)

        self._register_callback(callback_name)

        expected_seq = None

        # generate a sequence number and append it to the args
        if expect_seq:
            expected_seq = self._get_next_seq()
            args = (expected_seq,) + args

        if use_callback_rpc:
            # We need 'callback' to get us back to our random function
            callback_args = ('callback', callback_name, funcname) + args
        else:
            # If we have a specific callback, just make the RPC and the function should reply on its own
            callback_args = (funcname,) + args

        if message_type == MESSAGE_TYPE_DMCAST:
            # figure out how many addresses we can put in the packet.
            addrs, unprocessed_nodes = self._get_addrs_within_overhead(addrs, callback_args, node_limit)

        future = SnapFuture(message_type in [MESSAGE_TYPE_MCAST, MESSAGE_TYPE_DMCAST], expected_seq, addrs, callback_name)
        future_key = future.get_key()
        self._callback_futures[future_key] = future
        retry_failed_to_validate = {}

        # Send an RPC and wait for it to leave the system
        while retries >= 0:
            # Log the callback information.
            # Ex: "Callback RPC dmcast.some_function(arg1, arg2, arrg3) to ['001122'] timeout=3 retries=0"
            message_destination = addrs if message_type == MESSAGE_TYPE_RPC else message_type  # ex: 'aa0011', 'dmcast', 'mcast'
            log_responses = " to %s" % future.expected_responses if message_type == MESSAGE_TYPE_DMCAST else ""  # Only show expected responses for dmcast
            log.info("Callback RPC %s.%s%r%s timeout=%s retries=%s", message_destination, funcname, args, log_responses, timeout, retries)

            # RPC retries are handled separately from the response retries
            rpc_send_result = yield self._rpc_with_retries(message_type, funcname, callback_args,
                                                           future.expected_responses, ttl, dmcast_delay, group=group)

            if not rpc_send_result:
                # If we couldn't even send the message, there's no point in waiting for any replies.
                # Resolve the future with whatever we have outside of the loop
                log.error("Unable to send RPC after multiple attempts. Giving up and returning responses")
                break

            try:
                callback_result = yield with_timeout(timedelta(seconds=timeout), future)
            except TimeoutError:
                log.error("RPC timed out for %s %s%r (Retries=%d)", addrs if message_type == MESSAGE_TYPE_RPC else message_type, funcname, args, retries)
            else:
                callback_result, invalid_responses = self._validate_result(validate_func, callback_result)
                if invalid_responses:

                    if message_type == MESSAGE_TYPE_DMCAST or message_type == MESSAGE_TYPE_MCAST:
                        for k, v in invalid_responses.iteritems():
                            retry_failed_to_validate[k] = v

                    self._callback_futures.pop(future_key, None)
                    self._snap.local_dispatcher.funcs.pop(callback_name, None)

                    self._register_callback(callback_name)

                    future = SnapFuture(message_type in [MESSAGE_TYPE_MCAST, MESSAGE_TYPE_DMCAST], expected_seq, addrs,
                                        callback_name)
                    future_key = future.get_key()
                    self._callback_futures[future_key] = future
                else:
                    # If we didn't have a timeout error, break out of the loop
                    # Map from address to the invalid response tuple. In the responses dict, these will be set to None as if they did not respond.
                    callback_result, invalid_responses = self.get_results_with_result_type(result_type, callback_result)

                    if invalid_responses:
                        log.error("RPC received invalid responses for %s %s%r: %r",
                                  addrs if message_type == MESSAGE_TYPE_RPC else message_type,
                                  funcname, args, invalid_responses)

                    if message_type == MESSAGE_TYPE_DMCAST:
                        # dmcast messages further wrap the responses to include unprocessed nodes
                        callback_result = MultiResult(callback_result, unprocessed_nodes, invalid_responses)

                    raise Return(callback_result)

            retries -= 1

        # If we reached this point, we're out of retries, get rid of this future, we don't need it
        self._callback_futures.pop(future_key, None)
        self._snap.local_dispatcher.funcs.pop(callback_name, None)

        if message_type == MESSAGE_TYPE_DMCAST or message_type == MESSAGE_TYPE_MCAST:
            # If we're multicast or dmcast, we can still return if we heard from anyone, even if we timed out and never heard from everybody
            future_responses, failed_to_validate = self._validate_result(validate_func, future.responses)
            future_responses, failed_to_tuple = self.get_results_with_result_type(result_type, future_responses)
            response = {a: None for a in addrs}
            response.update(future_responses)

            invalid_responses = {}
            invalid_responses.update(failed_to_tuple)
            invalid_responses.update(failed_to_validate)
            invalid_responses.update(retry_failed_to_validate)

            if invalid_responses:
                log.error("RPC received invalid responses for %s %s%r: %r",
                          addrs if message_type == MESSAGE_TYPE_RPC else message_type,
                          funcname, args, invalid_responses)

            if message_type == MESSAGE_TYPE_DMCAST:
                response = MultiResult(response, unprocessed_nodes, invalid_responses)

            raise Return(response)
        else:
            # If we weren't a multicast or dmcast message, the future resolves to None at this point to indicate no response
            pass

    @staticmethod
    def _validate_result(validate_func, callback_result):
        """Runs the callback result against the validation function.

        Args:
            validate_func (function): The function to apply to each result of the callback result. This function should
                take one RPC response tuple and return True or False.
            callback_result (tuple | dict): The result to verify against the callback function. It will be a dict for
                mcast and dmcast responses and a tuple for unicast rpc responses.

        Returns:
            tuple(dict | tuple, dict): Result of the validation into (valid, invalid). If callback_result was a tuple,
                this returns it in the first index; otherwise this returns it in the second index. If callback_result
                was a dict, the valid results are placed in a dict in the first index, and the remaining in second.
        """
        if not callable(validate_func):
            return callback_result, {}

        if isinstance(callback_result, dict):
            # do dict thing
            new_callback_result = {}
            invalid_result = {}
            for k, v in callback_result.iteritems():
                try:
                    if validate_func(v):
                        new_callback_result[k] = v
                    else:
                        invalid_result[k] = v
                except Exception:
                    log.exception("Validate function for result %s=>%s raised an exception", k, v)
                    invalid_result[k] = v

            return new_callback_result, invalid_result
        else:
            try:
                if validate_func(callback_result):
                    return callback_result, {}
                else:
                    return {}, callback_result
            except Exception:
                log.exception("Validate function for result %s raised an exception", callback_result)
                return {}, callback_result

    @staticmethod
    def _get_addrs_within_overhead(addrs, args, node_limit=None):
        """Return the addresses that will fit into a given dmcast packet and those that will be unprocessed.

        Args:
            addrs (list[str]): List of addresses (ex: ['aabbcc'])
            args (list): Arguments passed to the function.  The function name is the first argument.
            node_limit (int): Number of nodes to use as a maximum
                None if we want to use as many as possible

        Returns:
            list, list: Two lists containing nodes that will and will not fit in the rpc, respectively
        """
        # Find out how many addresses will fit
        func_name = args[0]
        func_args = args[1:]
        max_addrs = get_max_dmcast_targets(func_name, *func_args)

        if max_addrs == 0 and len(addrs) != 0:
            # We have a problem: The caller specified targets, but we couldn't fit *any* of them
            # in the packet.
            raise CouldNotFitTargetsError(addrs, func_name, func_args)

        # Now, take the smaller of the node_limit and the addrs that will actually fit
        if node_limit is not None:
            max_addrs = min(max_addrs, node_limit)
        if max_addrs < len(addrs):
            log_rpc.debug("Only %d of %d addrs will fit in the next DMcast packet", max_addrs, len(addrs))

        unprocessed_nodes = addrs[max_addrs:]  # List of nodes that wouldn't fit in this message
        processed_nodes = addrs[:max_addrs]  # List of nodes that will fit into this message

        return processed_nodes, unprocessed_nodes

    def callback_rpc(self, addr, funcname, *args, **kwargs):
        """Send a callback RPC with automatic callback, timeout, and retry handling.

        Args:
            addr (str): SNAP address (Ex: "aabbcc", must be lowercase)
            funcname (str): Name of function to call with RPC
            *args : Arguments to pass to the node
            **kwargs : Keyword arguments for SNAPconnect. View Keyword Arguments for specific kwargs expected.

        Keyword Args:
            callback_name (str): If None, send the RPC with 'callback' to an automatically assigned callback value.
                If set to a value, listen for an automatic response from the RPC with the given value. Default : None
            result_type (namedtuple): Named tuple describing the format of the returned value. Default : None
            retries (int): Number of times to retry the message after a timeout. Default : 3
            timeout (float): Number of seconds to wait before timing out. Default : 2
            validate_func (function): The function to apply to each result of the callback. This function should
                take one RPC response tuple and return True or False.

        Raises:
            TypeError: Raised if an invalid keyword argument is provided.

        Returns:
            Future | None: A future that resolves to the return value of the function or None if we don't get a result
        """

        addr = validate_addresses(addr)
        valid_rpc_kwargs = ('timeout', 'retries', 'result_type', 'callback_name', 'validate_func')
        extra_kwargs = [key for key in kwargs if key not in valid_rpc_kwargs]

        if extra_kwargs:
            raise TypeError('callback_rpc() got an unexpected keyword argument %s' % extra_kwargs)

        return self._callback_message(addr, funcname, *args, message_type=MESSAGE_TYPE_RPC, **kwargs)

    def callback_mcast_rpc(self, addrs, funcname, *args, **kwargs):
        """Send a callback multicast RPC with automatic callback, timeout, and retry handling.

        This waits for every address to respond. If we're out of timeouts and retries, then it returns None.

        Args:
            addrs (str | list[str]): Single SNAP address or list of SNAP addresses we expect to respond
                (Ex: ["aabbcc", "ccddee"], must be lowercase).
                If blank/empty ('' or []), any responding node will be treated as if it were expected.
            funcname (str): Name of function to call with RPC
            *args : Arguments to pass to the node
            **kwargs : Keyword arguments for SNAPconnect. View Keyword Arguments for specific kwargs expected.

        Keyword Args:
            callback_name (str): If None, send the RPC with 'callback' to an automatically assigned callback value.
                If set to a value, listen for an automatic response from the RPC with the given value. Default : None
            group (int): Multicast Group. If not provided or None, use default group. Default : None
            result_type (namedtuple): Named tuple describing the format of the returned value. Default : None
            retries (int): Number of times to retry the message after a timeout. Default : 3
            timeout (float): Number of seconds to wait before timing out. Default : 2
            ttl (int): Number of hops. If not provided or None, use default ttl. Default : None
            validate_func (function): The function to apply to each result of the callback. This function should
                take one RPC response tuple and return True or False.

        Raises:
            TypeError: Raised if an invalid keyword argument is provided.

        Returns:
            Future | None: A future that resolves to the return value of the functions as a dictionary.
            None if we don't get a result.

        Note:
            * ``callback_mcast_rpc`` uses :meth:`~SnapConnectFutures.get_default_mcast_group` to obtain the default mcast group *unless overridden* via kwargs
            * ``callback_mcast_rpc`` uses :meth:`~SnapConnectFutures.get_default_mcast_ttl` to obtain the default mcast ttl *unless overridden* via kwargs
        """

        addrs = validate_addresses(addrs)
        valid_mcast_kwargs = ('timeout', 'retries', 'result_type', 'callback_name', 'ttl', 'group', 'validate_func')
        extra_kwargs = [key for key in kwargs if key not in valid_mcast_kwargs]

        if extra_kwargs:
            raise TypeError('callback_mcast_rpc() got an unexpected keyword argument %s' % extra_kwargs)

        # Unify: Treat empty string of addrs like empty list addrs
        if addrs == '':
            addrs = []

        if isinstance(addrs, basestring):
            addrs = [addrs]
        return self._callback_message(addrs, funcname, *args, message_type=MESSAGE_TYPE_MCAST, **kwargs)

    @staticmethod
    def _sort_by_attempts_remaining(attempts_remaining):
        """Return a sorted list of addresses with the most attempts remaining first

        Args:
            attempts_remaining (dict[str, int]): Dictionary addresses mapped to attempts remaining

        Returns:
            list: Sorted list of addresses with the most attempts remaining first
        """
        sorted_list = attempts_remaining.items()
        sorted_list.sort(key=lambda tup: tup[1], reverse=True)
        return [tup[0] for tup in sorted_list]

    @coroutine
    def callback_dmcast_rpc(self, addrs, funcname, *args, **kwargs):
        """Send a callback directed multicast RPC with automatic callback, timeout, and retry handling.

        This waits for every address to respond. If we're out of timeouts and retries, then it returns None.

        Example::

            "aabbcc": ("response1"), # This one responded with one argument
            "bbccdd": (10, 20, 30), # This one responded with 3 arguments
            "ddeeff": None,  # We didn't hear a response from this one
            ["aabb11", "bbcc22"]  # We couldn't fit these 2

        Args:
            addrs (str | list[str]): List of SNAP addresses we expect to respond
                (Ex: ["aabbcc", "ccddee"], must be lowercase).
                If blank/empty ('' or []), any responding node will be treated as if it were expected.
            funcname (str): Name of function to call with RPC
            *args : Arguments to pass to the node
            **kwargs : Keyword arguments for SNAPconnect. View Keyword Arguments for specific kwargs expected.

        Keyword Args:
            callback_name (str): If None, send the RPC with 'callback' to an automatically assigned callback value.
                If set to a value, listen for an automatic response from the RPC with the given value. Default : None
            dmcast_delay (int): Millisecond delay between dmcast responses.
                If not provided or None, use default delay. Default : None
            expect_seq (bool): True if a sequence number will be appended to the RPC args.
                Used in most lighting RPCs. Default : False
            group (int): Multicast Group. If not provided or None, use default group. Default : None
            node_limit (int | None): The max number of nodes allowed in a dmcast.
                None if we should fit as many as possible. Default : None
            result_type (namedtuple): Named tuple describing the format of the returned value. Default : None
            retries (int): Number of times to retry the message after a timeout. Default : 3
            timeout (float): Number of seconds to wait before timing out. Default : 2
            ttl (int): Number of hops. If not provided or None, use default ttl. Default : None
            validate_func (function): The function to apply to each result of the callback. This function should
                take one RPC response tuple and return True or False.

        Raises:
            TypeError: Raised if an invalid keyword argument is provided.

        Returns:
            Future | None: A future that resolves to the return value of the functions as a dictionary.
            None if we don't get a result.

        Other Parameters:
            chunk_results (bool): (Internal Use Only) If True, return the results in chunks with unprocessed nodes.
                If False, run until we have every single result and then return. Default : False
            previous_chunk (MultiResult): (Internal Use Only) Used to continue processing data when chunk_results is
                set to True. Used internally when chunk_results is set to True. Default : None

        Note:
            * ``callback_dmcast_rpc`` uses :meth:`~SnapConnectFutures.get_defalt_dmcast_delay` to obtain the default dmcast delay *unless overridden* via kwards
            * ``callback_dmcast_rpc`` uses :meth:`~SnapConnectFutures.get_default_mcast_group` to obtain the default mcast group *unless overridden* via kwargs
            * ``callback_dmcast_rpc`` uses :meth:`~SnapConnectFutures.get_default_mcast_ttl` to obtain the default mcast ttl *unless overridden* via kwargs
        """

        addrs = validate_addresses(addrs)
        # NOTE: This uses the 2.6 builtin dmcast_rpc - you must be using correct snapconnect / snaplib for this to work.
        valid_dmcast_kwargs = ('timeout', 'retries', 'result_type', 'callback_name', 'dmcast_delay', 'node_limit',
                               'ttl', 'expect_seq', 'chunk_results', 'previous_chunk', 'group', 'validate_func')
        extra_kwargs = [key for key in kwargs if key not in valid_dmcast_kwargs]

        if extra_kwargs:
            raise TypeError('callback_dmcast_rpc() got an unexpected keyword argument %s' % extra_kwargs)

        retries = kwargs.pop('retries', DEFAULT_RETRIES)
        previous_chunk = kwargs.pop('previous_chunk', None)
        chunk_results = kwargs.pop('chunk_results', False)
        kwargs['expect_seq'] = kwargs.get('expect_seq', False)

        # Unify: Treat empty list addrs like empty string addrs
        # Yes, this is the inverse of what we do for callback_mcast_rpc
        if addrs == []:
            addrs = ''

        if isinstance(addrs, basestring):
            addrs = [addrs]

        if previous_chunk is None:
            # Map from address to attempts remaining. OrderedDict preserves passed in order.
            # Set the initial number of attempts allowed for each node
            attempts_remaining = OrderedDict([(addr, retries + 1) for addr in addrs])
            current_results = {}  # Map from addr to the response object
            current_invalid_results = {}
        else:
            attempts_remaining = previous_chunk.attempts_remaining
            current_results = previous_chunk.responses
            current_invalid_results = previous_chunk.invalid_responses

        multi_response = None

        while attempts_remaining:
            addrs = self._sort_by_attempts_remaining(attempts_remaining)

            # Because we're managing retries in this method, set them to 0 in _callback_message
            multi_response = yield self._callback_message(addrs, funcname, *args, retries=0,
                                                          message_type=MESSAGE_TYPE_DMCAST, **kwargs)

            current_results.update(multi_response.responses)
            current_invalid_results.update(multi_response.invalid_responses)

            # Remove any successful nodes from attempts_remaining
            for addr, result in multi_response.responses.iteritems():
                if result is None and addr not in multi_response.invalid_responses:
                    # This is a failed attempt
                    attempts_remaining[addr] -= 1

                    if attempts_remaining[addr] == 0:
                        # Remove this from the list, we're done with it
                        attempts_remaining.pop(addr, 0)
                else:
                    # Remove this from the list, we heard from the node or it was invalid and we do not need to retry it
                    attempts_remaining.pop(addr, 0)

            if chunk_results:
                # If we're returning in chunks, we don't want to continue in this loop
                # Instead, we attach our state to MutliResult object and allow it to call a special .retry() function
                multi_response.chunk_responses = multi_response.responses.copy()  # Most recent responses are the chunk responses
                multi_response.attempts_remaining = attempts_remaining

                def retry_function():
                    return self.callback_dmcast_rpc(addrs, funcname, *args, retries=0,
                                                    chunk_results=True, previous_chunk=multi_response, **kwargs)

                multi_response.set_retry(retry_function)
                break  # Do not continue looping when chunking the results

        if multi_response:
            # Roll our most recent responses and failures into the totals.
            multi_response.responses.update(current_results)
            multi_response.invalid_responses.update(current_invalid_results)

            # KLUDGE: Clean up if we were asked to accept all responses...
            if addrs == ['']:
                # The {'': None} response is spurious...
                multi_response.responses.pop('', None)

            raise Return(multi_response)

    def open_serial(self, serial_type, port, forward_groups=None, baudrate=None, timeout=2):
        """Return a future that resolves to the bridge address when a serial connection is opened.

        Args:
            serial_type (int): The type of serial interface to open
            port (int | str): Port to connect to for the given serial type
            forward_groups (int): Multi-cast groups to forward through. Defaults to the NV parameter default.
            baudrate (int): The baudrate for the serial connection. Defaults to SNAPconnect's default.
            timeout (float): Time in seconds before the open attempt has failed

        Raises:
            TimeoutError: Serial connection did not open before timeout.

        Returns:
            Future: A future that resolves to the bridge address when a serial connection is opened.

        See Also:
            * :ref:`ConstantsUsedWithEncryption` for serial_type constants
        """
        return self.serial_manager.serial_operation(is_open=True, serial_type=serial_type, port=port,
                                                    forward_groups=forward_groups, baudrate=baudrate, timeout=timeout)

    def close_serial(self, serial_type, port, timeout=2):
        """Return a future that resolves to True when a serial connection is closed

        Args:
            serial_type (int): The type of serial interface to close
            port (int | str): Port to close for the given serial type
            timeout (float): Time in seconds before the close has failed

        Raises:
            TimeoutError: Serial connection did not close before timeout.

        Returns:
            Future: A future that resolves to True when a serial connection is closed

        See Also:
            * :ref:`ConstantsUsedWithEncryption` for serial_type constants
        """
        return self.serial_manager.serial_operation(is_open=False, serial_type=serial_type, port=port, timeout=timeout)

    def upgrade_script(self, addr, spy_file_path, timeout=60):
        """Return a future that resolves to True if the script upgrade was successful or False if it was not.

        Args:
            addr (str): SNAP address (Ex: "aabbcc", must be lowercase)
            spy_file_path (str): Full path to SNAPpy SPY file
            timeout (float): Time in seconds before an upgrade has failed

        Raises:
            TimeoutError: Unable to open script before timeout.

        Returns:
            Future: A future that resolves to True if the script upgrade was successful or False if it was not.
        """
        addr = validate_addresses(addr)
        return self.upgrade_manager.upgrade_script(addr, spy_file_path, timeout)

    def upgrade_firmware(self, addr, sfi_file_path, timeout=60):
        """Return a future that resolves to True if the firmware upgrade was successful or False if it was not.

        Args:
            addr (str): SNAP address (Ex: "aabbcc", must be lowercase)
            sfi_file_path (str): Full path to SNAP Firmware Image file
            timeout (float): Time in seconds before an upgrade has failed

        Raises:
            TimeoutError: Upgrade firmware doesn't complete before timeout.

        Returns:
            Future: A future that resolves to True if the firmware upgrade was successful or False if it was not.
        """
        addr = validate_addresses(addr)
        return self.upgrade_manager.upgrade_firmware(addr, sfi_file_path, timeout)
